from __future__ import annotations

from abc import ABC, abstractmethod
from math import inf
from sys import platform
from typing import TYPE_CHECKING, Any, Dict, Iterable, NewType, Optional, Sequence, Tuple, TypeVar

import numpy as np
from attrs import field, frozen
from numpy.typing import NDArray

try:
    from rtamt import Language, Semantics, StlDenseTimeSpecification, StlDiscreteTimeSpecification
except ImportError:
    _has_rtamt = False
else:
    _has_rtamt = True

from .core import Specification, SpecificationError

try:
    from .parser import parse
except:
    _can_parse = False
else:
    _can_parse = True

try:
    from taliro import tptaliro as tp
except ImportError:
    _has_tptaliro = False
else:
    _has_tptaliro = True

try:
    from .parser import SpecificationSyntaxError, TemporalLogic, translate
except ImportError:
    _can_translate = False
else:
    _can_translate = True

if TYPE_CHECKING:
    from taliro.tptaliro import AdjacencyList, Guard, GuardMap, HyDist

StateT = TypeVar("StateT")
PredicateName = str
ColumnT = int
PredicateColumnMap = Dict[PredicateName, ColumnT]


class StlSpecificationException(Exception):
    pass


_Times = NewType("_Times", NDArray[np.float64])
_States = NewType("_States", NDArray[np.float64])
_Parsed = Tuple[_Times, _States]


def _parse_times_states(times: Sequence[float], states: Sequence[Sequence[float]]) -> _Parsed:
    times_ = _Times(np.array(times, dtype=np.float64))
    states_ = _States(np.array(states, dtype=np.float64))

    if times_.ndim != 1:
        raise StlSpecificationException("Times must be a 1-D vector")

    if states_.ndim != 2:
        raise StlSpecificationException("States must be a 2-D matrix")

    if states_.shape[0] == times_.size:
        return (times_, states_.T)
    elif states_.shape[1] == times_.size:
        return (times_, states_)
    else:
        raise StlSpecificationException("States must have one dimension that maches times length")


class StlSpecification(Specification[Sequence[float], float], ABC):
    @abstractmethod
    def __init__(self, requirement: str, column_map: PredicateColumnMap):
        ...

    @property
    def failure_cost(self) -> float:
        return -inf


class TLTK(StlSpecification):
    """STL logic specification that uses TLTK to compute robustness values.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, column_map: PredicateColumnMap):
        if not _can_parse:
            if platform == "linux":
                raise RuntimeError(
                    "TLTK specifications require parsing functionality. Please refer to the documentation for how to enable parsing."
                )
            else:
                raise RuntimeError("TLTK specification is only available on Linux")

        tltk_obj = parse(phi, list(column_map.keys()))

        if tltk_obj is None:
            raise SpecificationError("could not parse formula")

        self.tltk_obj = tltk_obj
        self.column_map = column_map

    def evaluate(self, states: Sequence[Sequence[float]], times: Sequence[float]) -> float:
        times_, states_ = _parse_times_states(times, states)
        map_items = self.column_map.items()
        traces = {name: np.array(states_[column], dtype=np.float64) for name, column in map_items}
        timestamps = np.array(times_, dtype=np.float32)

        self.tltk_obj.reset()
        self.tltk_obj.eval_interval(traces, timestamps)

        return self.tltk_obj.robustness


class RTAMTDiscrete(StlSpecification):
    """STL logic specification that uses RTAMT discrete-time semantics to compute robustness.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, column_map: PredicateColumnMap):
        if "time" in column_map:
            raise SpecificationError("'time' cannot be used as a predicate name for RTAMT")

        self.phi = phi
        self.column_map = column_map

    def evaluate(self, states: Sequence[Sequence[float]], times: Sequence[float]) -> float:
        if not _has_rtamt:
            raise RuntimeError("RTAMT must be installed to use RTAMTDiscrete specification")

        times_, states_ = _parse_times_states(times, states)

        if times_.size < 2:
            raise RuntimeError("timestamps must have at least two samples to evaluate")

        spec = StlDiscreteTimeSpecification()

        for name in self.column_map:
            spec.declare_var(name, "float")

        period = times[1] - times[0]

        spec.set_sampling_period(round(period, 2), "s", 0.1)
        spec.spec = self.phi
        spec.parse()

        traces = {"time": list(times)}

        for name, column in self.column_map.items():
            traces[name] = list(states[column])

        robustness = spec.evaluate(traces)
        return robustness[0][1]


class RTAMTDense(StlSpecification):
    """STL logic specification that uses RTAMT dense-time semantics to compute robustness.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, column_map: PredicateColumnMap):
        self.phi = phi
        self.column_map = column_map

    def evaluate(self, states: Sequence[Sequence[float]], times: Sequence[float]) -> float:
        if not _has_rtamt:
            raise RuntimeError("RTAMT must be installed to use RTAMTDense specification")

        times_, states_ = _parse_times_states(times, states)
        spec = StlDenseTimeSpecification()

        for name in self.column_map:
            spec.declare_var(name, "float")

        spec.spec = self.phi
        spec.parse()

        map_items = self.column_map.items()
        traces = [
            (name, np.array([times_.tolist(), states_[column]]).T.tolist())
            for name, column in map_items
        ]

        robustness = spec.evaluate(*traces)
        return robustness[0][1]


@frozen(slots=True)
class TaliroPredicate:
    name: str = field(kw_only=True)
    A: NDArray[np.float_] = field(kw_only=True)
    b: NDArray[np.float_] = field(kw_only=True)
    l: Optional[NDArray[np.float_]] = field(default=None, kw_only=True)

    def as_dict(self) -> Dict[str, Any]:
        pred = {
            "name": self.name,
            "a": np.array(self.A, dtype=np.double, ndmin=2),
            "b": np.array(self.b, dtype=np.double, ndmin=2),
        }

        if self.l:
            pred["l"] = np.array(self.l, dtype=np.double, ndmin=2)

        return pred

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> TaliroPredicate:
        try:
            l = d["l"]
        except KeyError:
            l = None

        return cls(name=d["name"], A=d["a"], b=d["b"], l=l)


class TpTaliro(StlSpecification):
    """TPTL logic specification that uses TP-TaLiRo to compute robustness values.

    Attributes:
        phi: The specification requirement
        predicates: A set of Predicate(s) used in the requirement
    """

    def __init__(self, phi: str, predicate_map: Iterable[TaliroPredicate]):
        if not _has_tptaliro:
            if platform == "linux":
                raise RuntimeError("Py-TaLiRo must be installed to use TP-TaLiRo specification")
            else:
                raise RuntimeError("Py-TaLiRo is only available on Linux")

        if not _can_translate:
            raise RuntimeError("TP-TaLiRo specifications require translation functionality.")

        # translate STL to TPTL; else, assume valid TPTL
        try:
            self.spec = translate(phi, TemporalLogic.STL, TemporalLogic.TPTL)
        except SpecificationSyntaxError:
            self.spec = phi

        self.pmap = [user_dict.as_dict() for user_dict in predicate_map]

    def evaluate(self, states: Sequence[Sequence[float]], times: Sequence[float]) -> float:
        """Compute the euclidean-based robustness

        Attributes:
            states: State trajectories
            times: Timestamps
        """
        times_, states_ = _parse_times_states(times, states)

        robustness: HyDist = tp.tptaliro(
            spec=self.spec,
            preds=self.pmap,
            st=np.array(states_, dtype=np.double, ndmin=2),
            ts=np.array(times_, dtype=np.double, ndmin=2),
            lt=None,
            adj_list=None,
            guards=None,
        )

        return robustness["ds"]

    def hybrid(
        self,
        states: Sequence[Sequence[float]],
        times: Sequence[float],
        locations: Sequence[float],
        graph: AdjacencyList,
        guard_map: GuardMap,
    ) -> HyDist:
        """Compute hybrid-based robustness metric.

        A couple notes regarding the arguments: (1) `locations` is a trajectory
        of active states corresponding to a timestamp; (2) the `guard_map` must
        have an entry for _every_ edge defined in the `graph`. For example, if
        we have the following adjacency list:

        ```text
        A -> B
        A -> A
        ```

        Then, the guard map should contain an entry for the edge (A, B) and the
        edge (A, A). It is not required to add an entry for the edge (B, A) or
        (B, B) (i.e., the graph is nondeterministic).

        Arguments:
            states: State trajectories
            times: Timestamps
            locations: Location trajectories
            graph: Control location graph represented as an adjacency list
            guard_map: Transition criteria for each edge in the graph
        """

        def into_taliro_guard(constraint: Guard) -> Guard:
            return {
                "a": np.array(constraint["a"], dtype=np.double, ndmin=2),
                "b": np.array(constraint["b"], dtype=np.double, ndmin=2),
            }

        for guard, constraint in guard_map.items():
            guard_map[guard] = into_taliro_guard(constraint)

        robustness: HyDist = tp.tptaliro(
            spec=self.spec,
            preds=self.pmap,
            st=np.array(states, dtype=np.double, ndmin=2),
            ts=np.array(times, dtype=np.double, ndmin=2),
            lt=np.array(locations, dtype=np.double, ndmin=2),
            adj_list=graph,
            guards=guard_map,
        )

        return robustness



from __future__ import annotations

import enum
import os
import statistics as stats
from itertools import takewhile
from typing import Iterable, Optional, Sequence, Union

import numpy as np
from attr import frozen
from numpy.random import Generator, default_rng
from numpy.typing import NDArray
from scipy import optimize
from typing_extensions import Literal

from .core.interval import Interval
from .core.optimizer import ObjectiveFn, Optimizer
from .core.sample import Sample

Samples = Sequence[Sample]
Bounds = Sequence[Interval]


class Behavior(enum.IntEnum):
    """Behavior when falsifying case for system is encountered.

    Attributes:
        FALSIFICATION: Stop searching when the first falsifying case is encountered
        MINIMIZATION: Continue searching after encountering a falsifying case until iteration
                      budget is exhausted
    """

    FALSIFICATION = enum.auto()
    MINIMIZATION = enum.auto()


def _sample_uniform(bounds: Bounds, rng: Generator) -> Sample:
    return Sample([rng.uniform(bound.lower, bound.upper) for bound in bounds])


def _minimize(samples: Samples, func: ObjectiveFn[float], nprocs: Optional[int]) -> Iterable[float]:
    if nprocs is None:
        return func.eval_samples(samples)
    else:
        return func.eval_samples_parallel(samples, nprocs)


def _falsify(samples: Samples, func: ObjectiveFn[float]) -> Iterable[float]:
    costs = map(func.eval_sample, samples)
    return takewhile(lambda c: c >= 0, costs)


@frozen(slots=True)
class UniformRandomResult:
    """Data class that represents the result of a uniform random optimization.

    Attributes:
        average_cost: The average cost of all the samples selected.
    """

    average_cost: float


class UniformRandom(Optimizer[float, UniformRandomResult]):
    """Optimizer that implements the uniform random optimization technique.

    This optimizer picks samples randomly from the search space until the budget is exhausted.

    Args:
        parallelization: Value that indicates how many processes to use when evaluating each
                            sample using the cost function. Acceptable values are a number,
                            "cores", or None

    Attributes:
        processes: The number of processes to use when evaluating the samples.
    """

    def __init__(
        self,
        parallelization: Union[Literal["cores"], int, None] = None,
        behavior: Behavior = Behavior.FALSIFICATION,
    ):
        if isinstance(parallelization, int):
            self.processes: Optional[int] = parallelization
        elif parallelization == "cores":
            self.processes = os.cpu_count()
        else:
            self.processes = None

        self.behavior = behavior

    def optimize(
        self, func: ObjectiveFn[float], bounds: Bounds, budget: int, seed: int
    ) -> UniformRandomResult:
        rng = default_rng(seed)
        samples = [_sample_uniform(bounds, rng) for _ in range(budget)]

        if self.behavior is Behavior.MINIMIZATION:
            costs = _minimize(samples, func, self.processes)
        else:
            costs = _falsify(samples, func)

        average_cost = stats.mean(costs)

        return UniformRandomResult(average_cost)


@frozen(slots=True)
class DualAnnealingResult:
    """Data class representing the result of a dual annealing optimization.

    Attributes:
        jacobian_value: The value of the cost function jacobian at the minimum cost discovered
        jacobian_evals: Number of times the jacobian of the cost function was evaluated
        hessian_value: The value of the cost function hessian as the minimum cost discovered
        hessian_evals: Number of times the hessian of the cost function was evaluated
    """

    jacobian_value: Optional[NDArray[np.float_]]
    jacobian_evals: int
    hessian_value: Optional[NDArray[np.float_]]
    hessian_evals: int


class DualAnnealing(Optimizer[float, DualAnnealingResult]):
    """Optimizer that implements the simulated annealing optimization technique.

    The simulated annealing implementation is provided by the SciPy library dual_annealing function
    with the no_local_search parameter set to True.
    """

    def __init__(self, behavior: Behavior = Behavior.FALSIFICATION):
        self.behavior = behavior

    def optimize(
        self, func: ObjectiveFn[float], bounds: Bounds, budget: int, seed: int
    ) -> DualAnnealingResult:
        def listener(sample: NDArray[np.float_], robustness: float, ctx: Literal[-1, 0, 1]) -> bool:
            if robustness < 0 and self.behavior is Behavior.FALSIFICATION:
                return True

            return False

        result = optimize.dual_annealing(
            func=lambda x: func.eval_sample(Sample(x)),
            bounds=[bound.astuple() for bound in bounds],
            seed=seed,
            maxfun=budget,
            no_local_search=True,  # Disable local search, use only traditional generalized SA
            callback=listener,
        )

        try:
            jac: Optional[NDArray[np.float_]] = result.jac
            njev = result.njev
        except AttributeError:
            jac = None
            njev = 0

        try:
            hess: Optional[NDArray[np.float_]] = result.hess
            nhev = result.nhev
        except AttributeError:
            hess = None
            nhev = 0

        return DualAnnealingResult(jac, njev, hess, nhev)



# Generated from stlParser.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .stlParser import stlParser
else:
    from stlParser import stlParser

# This class defines a complete listener for a parse tree produced by stlParser.
class stlParserListener(ParseTreeListener):

    # Enter a parse tree produced by stlParser#stlSpecification.
    def enterStlSpecification(self, ctx:stlParser.StlSpecificationContext):
        pass

    # Exit a parse tree produced by stlParser#stlSpecification.
    def exitStlSpecification(self, ctx:stlParser.StlSpecificationContext):
        pass


    # Enter a parse tree produced by stlParser#predicateExpr.
    def enterPredicateExpr(self, ctx:stlParser.PredicateExprContext):
        pass

    # Exit a parse tree produced by stlParser#predicateExpr.
    def exitPredicateExpr(self, ctx:stlParser.PredicateExprContext):
        pass


    # Enter a parse tree produced by stlParser#opFutureExpr.
    def enterOpFutureExpr(self, ctx:stlParser.OpFutureExprContext):
        pass

    # Exit a parse tree produced by stlParser#opFutureExpr.
    def exitOpFutureExpr(self, ctx:stlParser.OpFutureExprContext):
        pass


    # Enter a parse tree produced by stlParser#parenPhiExpr.
    def enterParenPhiExpr(self, ctx:stlParser.ParenPhiExprContext):
        pass

    # Exit a parse tree produced by stlParser#parenPhiExpr.
    def exitParenPhiExpr(self, ctx:stlParser.ParenPhiExprContext):
        pass


    # Enter a parse tree produced by stlParser#opUntilExpr.
    def enterOpUntilExpr(self, ctx:stlParser.OpUntilExprContext):
        pass

    # Exit a parse tree produced by stlParser#opUntilExpr.
    def exitOpUntilExpr(self, ctx:stlParser.OpUntilExprContext):
        pass


    # Enter a parse tree produced by stlParser#opGloballyExpr.
    def enterOpGloballyExpr(self, ctx:stlParser.OpGloballyExprContext):
        pass

    # Exit a parse tree produced by stlParser#opGloballyExpr.
    def exitOpGloballyExpr(self, ctx:stlParser.OpGloballyExprContext):
        pass


    # Enter a parse tree produced by stlParser#opLogicalExpr.
    def enterOpLogicalExpr(self, ctx:stlParser.OpLogicalExprContext):
        pass

    # Exit a parse tree produced by stlParser#opLogicalExpr.
    def exitOpLogicalExpr(self, ctx:stlParser.OpLogicalExprContext):
        pass


    # Enter a parse tree produced by stlParser#opReleaseExpr.
    def enterOpReleaseExpr(self, ctx:stlParser.OpReleaseExprContext):
        pass

    # Exit a parse tree produced by stlParser#opReleaseExpr.
    def exitOpReleaseExpr(self, ctx:stlParser.OpReleaseExprContext):
        pass


    # Enter a parse tree produced by stlParser#opNextExpr.
    def enterOpNextExpr(self, ctx:stlParser.OpNextExprContext):
        pass

    # Exit a parse tree produced by stlParser#opNextExpr.
    def exitOpNextExpr(self, ctx:stlParser.OpNextExprContext):
        pass


    # Enter a parse tree produced by stlParser#opPropExpr.
    def enterOpPropExpr(self, ctx:stlParser.OpPropExprContext):
        pass

    # Exit a parse tree produced by stlParser#opPropExpr.
    def exitOpPropExpr(self, ctx:stlParser.OpPropExprContext):
        pass


    # Enter a parse tree produced by stlParser#opNegExpr.
    def enterOpNegExpr(self, ctx:stlParser.OpNegExprContext):
        pass

    # Exit a parse tree produced by stlParser#opNegExpr.
    def exitOpNegExpr(self, ctx:stlParser.OpNegExprContext):
        pass


    # Enter a parse tree produced by stlParser#predicate.
    def enterPredicate(self, ctx:stlParser.PredicateContext):
        pass

    # Exit a parse tree produced by stlParser#predicate.
    def exitPredicate(self, ctx:stlParser.PredicateContext):
        pass


    # Enter a parse tree produced by stlParser#interval.
    def enterInterval(self, ctx:stlParser.IntervalContext):
        pass

    # Exit a parse tree produced by stlParser#interval.
    def exitInterval(self, ctx:stlParser.IntervalContext):
        pass





from __future__ import annotations

import statistics as stats
from abc import abstractmethod
from typing import Any, Generic, Iterable, Optional, Protocol, Sequence, Tuple, TypeVar, cast

import numpy as np
from attr import frozen
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from .interval import Interval
from .layout import SampleLayout
from .sample import Sample
from .signal import Signal

ResultT = TypeVar("ResultT")
ExtraT = TypeVar("ExtraT")
CostT = TypeVar("CostT")


class Comparable(Protocol):
    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __lt__(self: ComparableT, other: ComparableT) -> bool:
        pass


ComparableT = TypeVar("ComparableT", bound=Comparable)


@frozen()
class TimingData:
    """Storage class for execution durations of different PSY-TaLiRo components.

    The durations stored in this class are for a single evaluation.

    Attributes:
        model: Run time of model component
        specification: Run time of specification component
    """

    model: float
    specification: float

    @property
    def total(self) -> float:
        """The total duration of all components."""

        return self.model + self.specification


@frozen()
class Evaluation(Generic[CostT, ExtraT]):
    """The result of applying the cost function to a sample.

    Attributes:
        cost: The result of using a specification to analyze the output of a model
        sample: The sample provided to the model
        extra: Additional data returned by the model
        timing: Execution durations of each component of the cost function
    """

    cost: CostT
    sample: Sample
    extra: ExtraT
    timing: TimingData


@frozen(slots=True)
class TimeStats:
    """Data class that represents the standard statistics of a set of durations."""

    durations: Iterable[float]

    @property
    def total_duration(self) -> float:
        """The sum of all durations."""

        return sum(self.durations)

    @property
    def avg_duration(self) -> float:
        """The average of all durations."""

        return stats.mean(self.durations)

    @property
    def max_duration(self) -> float:
        """The maximum duration."""

        return max(self.durations)

    @property
    def min_duration(self) -> float:
        """The maximum duration."""

        return min(self.durations)


@frozen(slots=True)
class Run(Generic[ResultT, CostT, ExtraT]):
    """Data class that represents one run of an optimizer.

    Attributes:
        result: The value returned by the optimizer
        history: List of Evaluation instances representing every cost function evaluation
        duration: Time spent by the optimizer
    """

    result: ResultT
    history: Sequence[Evaluation[CostT, ExtraT]]
    duration: float
    seed: int

    @property
    def fastest_eval(self) -> Evaluation[CostT, ExtraT]:
        """The evaluation with the lowest total duration."""

        return min(self.history, key=lambda e: e.timing.total)

    @property
    def slowest_eval(self) -> Evaluation[CostT, ExtraT]:
        """Evaluation with the longest total duration."""

        return max(self.history, key=lambda e: e.timing.total)

    @property
    def model_timing(self) -> TimeStats:
        """Time statistics for the model."""

        return TimeStats(iteration.timing.model for iteration in self.history)

    @property
    def specification_timing(self) -> TimeStats:
        """Time statistics of the specification."""

        return TimeStats(iteration.timing.specification for iteration in self.history)


def best_eval(run: Run[Any, ComparableT, ExtraT]) -> Evaluation[ComparableT, ExtraT]:
    return max(run.history, key=lambda e: e.cost)


def worst_eval(run: Run[Any, ComparableT, ExtraT]) -> Evaluation[ComparableT, ExtraT]:
    return min(run.history, key=lambda e: e.cost)


@frozen(slots=True)
class Result(Generic[ResultT, CostT, ExtraT]):
    """Data class that represents a set of successful runs of the optimizer.

    Attributes:
        runs: List of Run instances representing each optimization attempt
        options: Configuration class used to control Model, Specification and Optimizer behaviors
                 for each run
    """

    runs: Sequence[Run[ResultT, CostT, ExtraT]]
    interval: Interval
    seed: int
    processes: Optional[int]
    layout: SampleLayout

    def plot_signal(self, signal: Signal, step_size: float = 0.1) -> Tuple[Figure, Axes]:
        fig, ax = plt.subplots()
        times = np.arange(self.interval.lower, self.interval.upper, step_size, dtype=np.float64)
        values = signal.at_times(cast(Sequence[float], times.tolist()))

        ax.plot(times, values)

        return fig, ax


def best_run(result: Result[ResultT, ComparableT, ExtraT]) -> Run[ResultT, ComparableT, ExtraT]:
    return max(result.runs, key=best_eval)


def worst_run(result: Result[ResultT, ComparableT, ExtraT]) -> Run[ResultT, ComparableT, ExtraT]:
    return min(result.runs, key=worst_eval)



from __future__ import annotations

from math import cos
from typing import List, Sequence, cast

import numpy as np
from scipy.interpolate import PchipInterpolator, interp1d

from .core.signal import Signal, SignalFactory


class Pchip(Signal):
    def __init__(self, interp: PchipInterpolator):
        self.interp = interp

    def at_time(self, t: float) -> float:
        return float(self.interp(t))

    def at_times(self, ts: Sequence[float]) -> List[float]:
        return cast(List[float], self.interp(ts).tolist())


def pchip(times: Sequence[float], signal_values: Sequence[float]) -> Pchip:
    return Pchip(PchipInterpolator(times, signal_values))


class Piecewise(Signal):
    def __init__(self, interp: interp1d):
        self.interp = interp

    def at_time(self, t: float) -> float:
        return float(self.interp(t))

    def at_times(self, ts: Sequence[float]) -> List[float]:
        return cast(List[float], self.interp(ts).tolist())


def piecewise_linear(times: Sequence[float], signal_values: Sequence[float]) -> Piecewise:
    return Piecewise(interp1d(times, signal_values))


def piecewise_constant(times: Sequence[float], signal_values: Sequence[float]) -> Piecewise:
    return Piecewise(interp1d(times, signal_values, kind="zero", fill_value="extrapolate"))


class Delayed(Signal):
    def __init__(self, signal: Signal, cutoff: int):
        self.signal = signal
        self.cutoff = cutoff

    def at_time(self, t: float) -> float:
        if t < self.cutoff:
            return 0.0

        return self.signal.at_time(t)


def delayed(signal_factory: SignalFactory, *, delay: int) -> SignalFactory:
    def factory(times: Sequence[float], signal_values: Sequence[float]) -> Signal:
        stop_time = max(times)
        new_times = np.linspace(
            start=delay, stop=stop_time, num=len(signal_values), dtype=np.float64
        )
        signal = signal_factory(new_times.tolist(), signal_values)

        return Delayed(signal, delay)

    return factory


class Sequenced(Signal):
    def __init__(self, s1: Signal, s2: Signal, t_switch: int):
        self.s1 = s1
        self.s2 = s2
        self.t_switch = t_switch

    def at_time(self, t: float) -> float:
        return self.s1.at_time(t) if t < self.t_switch else self.s2.at_time(t)


def sequenced(factory1: SignalFactory, factory2: SignalFactory, *, t_switch: int) -> SignalFactory:
    def factory(times: Sequence[float], signal_values: Sequence[float]) -> Signal:
        signal1_indices = [i for i, t in enumerate(times) if t < t_switch]
        signal2_indices = [i for i, t in enumerate(times) if t >= t_switch]
        signal1 = factory1(
            [times[i] for i in signal1_indices], [signal_values[i] for i in signal1_indices]
        )
        signal2 = factory2(
            [times[i] for i in signal2_indices], [signal_values[i] for i in signal2_indices]
        )

        return Sequenced(signal1, signal2, t_switch)

    return factory


class Harmonic(Signal):
    class Component:
        def __init__(self, amplitude: float, frequency: float, phase: float):
            self.theta = amplitude
            self.omega = frequency
            self.phi = phase

        def at_time(self, time: float) -> float:
            return self.theta * cos(self.omega * time - self.phi)

    def __init__(self, bias: float, components: Sequence[Harmonic.Component]):
        self.bias = bias
        self.components = tuple(components)

    def at_time(self, time: float) -> float:
        return self.bias + sum(component.at_time(time) for component in self.components)


def harmonic(_: Sequence[float], values: Sequence[float]) -> Harmonic:
    if len(values[1:]) % 3 != 0:
        raise RuntimeError("Insufficient number of values to generate a harmonic signal")

    bias = values[0]
    component_params = [(values[i], values[i + 1], values[i + 2]) for i in range(1, len(values), 3)]
    components = [Harmonic.Component(amp, freq, phase) for amp, freq, phase in component_params]

    return Harmonic(bias, components)


class Clamped(Signal):
    def __init__(self, s: Signal, lo: float, hi: float):
        self._signal = s
        self._lo = lo
        self._hi = hi

    def at_time(self, time: float) -> float:
        return min(self._hi, max(self._lo, self._signal.at_time(time)))


def clamped(factory: SignalFactory, *, lo: float, hi: float) -> SignalFactory:
    def _factory(times: Sequence[float], values: Sequence[float]) -> Clamped:
        return Clamped(factory(times, values), lo, hi)

    return _factory



from __future__ import annotations

import logging
import time
from concurrent.futures import ProcessPoolExecutor
from itertools import islice
from typing import Any, Callable, Generic, Iterable, Iterator, Optional, Sequence, TypeVar, Union

from attr import Attribute, field, frozen
from attr.validators import deep_iterable, instance_of, optional
from numpy.random import Generator, default_rng

from .cost import CostFn, SpecificationOrFactory
from .interval import Interval
from .layout import SampleLayout
from .model import Model
from .optimizer import Optimizer
from .result import Result, Run

StateT = TypeVar("StateT")
CostT = TypeVar("CostT")
ResultT = TypeVar("ResultT")
ExtraT = TypeVar("ExtraT")

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@frozen()
class CostFnGenerator(Generic[StateT, CostT, ExtraT], Iterator[CostFn[StateT, CostT, ExtraT]]):
    """Iterator class that generates new instances of CostFn.

    Attributes:
        model: Model instance representing the system use for simulation
        spec: Specification instance for evaluating system trajectories
        options: Configuration object to control behavior of the model and specification
    """

    model: Model[StateT, ExtraT]
    specification: SpecificationOrFactory[StateT, CostT]
    interval: Interval
    layout: SampleLayout

    def __next__(self) -> CostFn[StateT, CostT, ExtraT]:
        return CostFn(
            self.model,
            self.specification,
            self.interval,
            self.layout,
        )


@frozen()
class Experiment(Generic[StateT, CostT, ResultT, ExtraT]):
    """Class that represents a single optimization attempt.

    Attributes:
        cost_fn: CostFn instance that is specific to this optimization attempt
        optimizer: Optimizer instance shared by all optimization attempts
        params: OptimizationParams instance specific to this optimization attempt
    """

    cost_fn: CostFn[StateT, CostT, ExtraT]
    optimizer: Optimizer[CostT, ResultT]
    bounds: Sequence[Interval]
    iterations: int
    seed: int

    def run(self) -> Run[ResultT, CostT, ExtraT]:
        """Execute this optimization attempt.

        Returns:
            A Run instance containing the result from the optimizer, the evaluation history of the
            cost function and the overall duration of the optimization attempt.
        """

        start_time = time.perf_counter()
        result = self.optimizer.optimize(self.cost_fn, self.bounds, self.iterations, self.seed)
        stop_time = time.perf_counter()
        duration = stop_time - start_time

        return Run(result, self.cost_fn.history, duration, self.seed)


@frozen()
class ExperimentGenerator(
    Generic[StateT, CostT, ResultT, ExtraT], Iterable[Experiment[StateT, CostT, ResultT, ExtraT]]
):
    """Generate experiments from a sequence of cost functions and a random number generator.

    Generated experiments will be created by iterating over the sequence of cost functions. For
    each cost function, a random number is generated by the random number generator which is used as
    the seed for the experiment.

    Attributes:
        cost_fn_generator: A generator that returns a sequence of cost functions
        optimizer: The optimizer to use for each experiment
        bounds: The set of intervals that represents the values to be generated by the optimizer
        iterations: The sample budget for each experiment
        rng: The random number generator used to produce a seed for each experiment
    """

    cost_fns: Iterable[CostFn[StateT, CostT, ExtraT]]
    optimizer: Optimizer[CostT, ResultT]
    bounds: Sequence[Interval]
    iterations: int
    rng: Generator

    def __iter__(self) -> Iterator[Experiment[StateT, CostT, ResultT, ExtraT]]:
        for cost_fn in self.cost_fns:
            yield Experiment(
                cost_fn,
                self.optimizer,
                self.bounds,
                self.iterations,
                self.rng.integers(0, 2**32 - 1),
            )


def _greater_than(bound: float) -> Callable[[Any, Attribute[Any], Any], None]:
    def validator(_: Any, attr: Attribute[Any], value: Any) -> None:
        if float(value) <= bound:
            raise ValueError(f"{attr.name} must be greater than {bound}")

    return validator


def _within_range(lower: float, upper: float) -> Callable[[Any, Attribute[Any], Any], None]:
    def validator(_: Any, attr: Attribute[Any], value: Any) -> None:
        if not lower <= value <= upper:
            raise ValueError(f"{attr.name} must be inside interval [{lower}, {upper}]")

    return validator


def _min_length(length: int) -> Callable[[Any, Attribute[Any], Any], None]:
    def validator(_: Any, attr: Attribute[Sequence[Any]], value: Sequence[Any]) -> None:
        if len(value) < length:
            raise ValueError(f"{attr.name} must have minimum length of {length}")

    return validator


def _subclass_of(
    class_t: Union[type, tuple[type, ...]]
) -> Callable[[Any, Attribute[Any], Any], None]:
    def validator(_: Any, attr: Attribute[Any], value: Any) -> None:
        if not issubclass(type(value), class_t):
            raise TypeError(f"Expected {attr.name} to have type {class_t}. Got {type(value)}")

    return validator


@frozen(auto_attribs=False)
class Scenario(Generic[StateT, CostT, ResultT, ExtraT]):
    """Class that represents a set of optimization attempts.

    Optimization attempts can be run sequentially or in parallel. When run in parallel, the same
    optimizer instance is shared between all processes.

    Attributes:
        model: Model instance representing the system to use for simulation
        specification: Specification instance representing the requirement
        options: Configuration object to control the behavior of the model, specification and
                 optimizer
    """

    model: Model[StateT, ExtraT] = field(validator=_subclass_of(Model))
    specification: SpecificationOrFactory[StateT, CostT] = field()
    runs: int = field(validator=[instance_of(int), _greater_than(0)])
    iterations: int = field(validator=[instance_of(int), _greater_than(0)])
    seed: int = field(validator=[instance_of(int), _within_range(0, 2**32 - 1)])
    processes: Optional[int] = field(validator=optional([instance_of(int), _greater_than(0)]))
    bounds: Sequence[Interval] = field(
        validator=deep_iterable(instance_of(Interval), iterable_validator=_min_length(1))
    )
    interval: Interval = field(validator=instance_of(Interval))
    layout: SampleLayout = field(validator=instance_of(SampleLayout))

    def run(self, optimizer: Optimizer[CostT, ResultT]) -> Result[ResultT, CostT, ExtraT]:
        """Execute all optimization attempts given an optimizer.

        Args:
            optimizer: Optimizer to use for all optimization attempts.

        Returns:
            A list of Run instances representing the outcomes of each optimization attempt.
        """

        logger.debug("Beginning experiment")
        logger.debug(f"{self.runs} runs of {self.iterations} iterations")
        logger.debug(f"Initial seed: {self.seed}")
        logger.debug(f"Parallelization: {self.processes}")

        rng = default_rng(self.seed)
        cost_fns = CostFnGenerator(
            self.model,
            self.specification,
            self.interval,
            self.layout,
        )
        experiment_gen = ExperimentGenerator(cost_fns, optimizer, self.bounds, self.iterations, rng)
        experiments = islice(experiment_gen, self.runs)

        if self.processes is not None:
            with ProcessPoolExecutor(self.processes) as executor:
                futures: Iterable[Run[ResultT, CostT, ExtraT]] = executor.map(
                    Experiment.run, experiments
                )
                runs = list(futures)
        else:
            runs = [experiment.run() for experiment in experiments]

        return Result(runs, self.interval, self.seed, self.processes, self.layout)


__all__ = ["Scenario"]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Sequence

from typing_extensions import Protocol


class Signal(ABC):
    """Representation of a time-varying input to a system."""

    @abstractmethod
    def at_time(self, time: float) -> float:
        """Get the value of the signal at the specified time."""
        ...

    def at_times(self, times: Sequence[float]) -> List[float]:
        """Get the value of the signal at each specified time."""

        return [self.at_time(time) for time in times]


class SignalFactory(Protocol):
    """Construct a Signal using the provided times and signal values."""

    def __call__(self, __times: Sequence[float], __values: Sequence[float]) -> Signal:
        ...



from __future__ import annotations

from enum import IntEnum, auto
from typing import Any, Dict, Optional, Sequence, Union

import tltk_mtl as mtl
from antlr4.CommonTokenStream import CommonTokenStream
from antlr4.InputStream import InputStream

from .stlLexer import stlLexer as Lexer
from .stlParser import stlParser as Parser
from .stlParserVisitor import stlParserVisitor as Visitor

PredicateNameSeq = Sequence[str]
PredicateDict = Dict[str, mtl.Predicate]
Predicates = Union[PredicateNameSeq, PredicateDict]
TltkObject = Union[
    mtl.And,
    mtl.Finally,
    mtl.Global,
    mtl.Implication,
    mtl.Next,
    mtl.Not,
    mtl.Or,
    mtl.Predicate,
    mtl.Until,
]


class TemporalLogic(IntEnum):
    """Currently supported Temporal Logic (TL) specifications.

    Attributes:
       STL: Signal Temporal Logic
       TPTL: Timed Propositional Temporal Logic
    """

    STL = auto()
    TPTL = auto()

    def __lt__(self, other: int) -> Any:
        if not isinstance(other, TemporalLogic):
            return NotImplemented

        return self.value < other.value

    def __str__(self) -> str:
        return self.name


def parse(formula: str, predicates: Predicates, mode: str = "cpu") -> Optional[TltkObject]:
    """TLTk parser parses a specification requirement into an equivalent TLTk structure

    Attributes:
        formula: The formal specification requirement
        predicates: The set of Predicate(s) used in the requirement
        mode: The TLTk computation mode
    """
    input_stream = InputStream(formula)

    lexer = Lexer(input_stream)
    stream = CommonTokenStream(lexer)

    parser = Parser(stream)
    tree = parser.stlSpecification()
    visitor = Visitor(lexer, predicates, mode)

    return visitor.visit(tree)  # type: ignore


class SpecificationSyntaxError(Exception):
    """Exception to represent a syntax error in a specification."""

    pass


def _stl_to_tptl(stream: InputStream) -> str:
    """Translate STL to TPTL.

    Arguments:
        stream: ANTLR input stream.

    Returns:
        An equivalently translated formula.
    """

    from .stlLexer import stlLexer as Lexer
    from .stlParser import stlParser as Parser
    from .stlTptlParserVisitorTranslator import stlTptlParserVisitorTranslator as TptlVisitor

    try:
        lexer = Lexer(stream)
        stream = CommonTokenStream(lexer)

        parser = Parser(stream)
        tree = parser.stlSpecification()
        visitor = TptlVisitor(lexer)

        return visitor.visit(tree)  # type: ignore
    except:
        raise SpecificationSyntaxError("STL formula has a syntax error")


class UnsupportedTranslationError(Exception):
    """An exception thrown when an invalid translation scheme is selected."""

    pass


def translate(formula: str, source: TemporalLogic, target: TemporalLogic) -> str:
    """Translate a source TL to a target TL.

    Arguments:
        formula: The source TL specification.
        source: The source Temporal Logic formula to translate from.
        target: The target Temporal Logic formula to translate to.

    Returns:
        An equivalently translated formula.
    """

    stream = InputStream(formula)

    if source == TemporalLogic.STL and target == TemporalLogic.TPTL:
        return _stl_to_tptl(stream)
    elif source == target:
        return formula
    else:
        raise UnsupportedTranslationError(f"cannot translate from {source} to {target}")



# Generated from stlParser.g4 by ANTLR 4.5.1
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .stlParser import stlParser
else:
    from stlParser import stlParser


class stlTptlParserVisitorTranslator(ParseTreeVisitor):
    """Traverse the generated ANTLR parse tree and translate STL to TPTL."""

    def __init__(self, lexer):
        self.lexer = lexer
        self.freeze = 0

    def freezeTime(self, freeze, lower, upper, subformulas, token_type):
        translation = ""
        if token_type == self.lexer.FUTUREOP:
            translation = (
                f"@Var_t{freeze} <>"
                f"(({{ Var_t{freeze} >= {str(lower)} }} /\\ {{ Var_t{freeze} <= {str(upper)} }}) "
                f"/\\ {subformulas[0]})"
            )
        elif token_type == self.lexer.GLOBALLYOP:
            translation = (
                f"@Var_t{freeze} []"
                f"(({{ Var_t{freeze} >= {str(lower)} }} /\\ {{ Var_t{freeze} <= {str(upper)} }}) "
                f"-> {subformulas[0]})"
            )
        elif token_type == self.lexer.NEXTOP:
            translation = (
                f"@Var_t{freeze} X"
                f"(({{ Var_t{freeze} >= {str(lower)} }} /\\ {{ Var_t{freeze} <= {str(upper)} }}) "
                f"/\\ {subformulas[0]})"
            )
        elif token_type == self.lexer.UNTILOP:
            translation = (
                f"@Var_t{freeze} ( {subformulas[0]} U "
                f"(({{ Var_t{freeze} >= {str(lower)} }} /\\ {{ Var_t{freeze} <= {str(upper)} }}) "
                f"/\\ {subformulas[1]}))"
            )
        elif token_type == self.lexer.RELEASEOP:
            translation = (
                f"!(@Var_t{freeze} (!{subformulas[0]} U "
                f"(({{ Var_t{freeze} >= {str(lower)} }} /\\ {{ Var_t{freeze} <= {str(upper)} }}) "
                f"/\\ !{subformulas[1]})))"
            )

        return translation

    # Visit a parse tree produced by stlParser#stlSpecification.
    def visitStlSpecification(self, ctx: stlParser.StlSpecificationContext) -> str:
        return self.visit(ctx.getRuleContext().getChild(0))

    # Visit a parse tree produced by stlParser#predicateExpr.
    def visitPredicateExpr(self, ctx: stlParser.PredicateExprContext) -> str:
        return self.visit(ctx.getRuleContext().getChild(0))

    # Visit a parse tree produced by stlParser#opFutureExpr.
    def visitOpFutureExpr(self, ctx: stlParser.OpFutureExprContext) -> str:
        child_count = ctx.getRuleContext().getChildCount()

        phi = ""
        if child_count == 3:
            index = self.freeze
            self.freeze += 1

            bounds = self.visit(ctx.getRuleContext().getChild(1))
            phi = self.visit(ctx.getRuleContext().getChild(2))
            token_type = ctx.getRuleContext().getChild(0).getSymbol().type

            return self.freezeTime(index, bounds[0], bounds[1], [phi], token_type)
        else:
            phi = self.visit(ctx.getRuleContext().getChild(1))
            return "<> " + phi

    # Visit a parse tree produced by stlParser#parenPhiExpr.
    def visitParenPhiExpr(self, ctx: stlParser.ParenPhiExprContext) -> str:
        return "( " + self.visit(ctx.getRuleContext().getChild(1)) + " )"

    # Visit a parse tree produced by stlParser#opUntilExpr.
    def visitOpUntilExpr(self, ctx: stlParser.OpUntilExprContext) -> str:
        child_count = ctx.getRuleContext().getChildCount()

        phi_1 = ""
        phi_2 = ""

        if child_count == 4:
            index = self.freeze
            self.freeze += 1

            phi_1 = self.visit(ctx.getRuleContext().getChild(0))
            phi_2 = self.visit(ctx.getRuleContext().getChild(3))

            bounds = self.visit(ctx.getRuleContext().getChild(2))
            token_type = ctx.getRuleContext().getChild(1).getSymbol().type

            return self.freezeTime(index, bounds[0], bounds[1], [phi_1, phi_2], token_type)
        else:
            phi_1 = self.visit(ctx.getRuleContext().getChild(0))
            phi_2 = self.visit(ctx.getRuleContext().getChild(2))

            return phi_1 + " U " + phi_2

    # Visit a parse tree produced by stlParser#opGloballyExpr.
    def visitOpGloballyExpr(self, ctx: stlParser.OpGloballyExprContext) -> str:
        child_count = ctx.getRuleContext().getChildCount()

        phi = ""
        if child_count == 3:
            index = self.freeze
            self.freeze += 1

            bounds = self.visit(ctx.getRuleContext().getChild(1))
            phi = self.visit(ctx.getRuleContext().getChild(2))
            token_type = ctx.getRuleContext().getChild(0).getSymbol().type

            return self.freezeTime(index, bounds[0], bounds[1], [phi], token_type)
        else:
            phi = self.visit(ctx.getRuleContext().getChild(1))
            return "[] " + phi

    # Visit a parse tree produced by stlParser#opLogicalExpr.
    def visitOpLogicalExpr(self, ctx: stlParser.OpLogicalExprContext) -> str:
        token_type = ctx.getRuleContext().getChild(1).getSymbol().type

        phi_1 = ""
        phi_2 = ""

        if token_type == self.lexer.ANDOP:
            # conjunction
            phi_1 = self.visit(ctx.getRuleContext().getChild(0))
            phi_2 = self.visit(ctx.getRuleContext().getChild(2))

            return phi_1 + r" /\ " + phi_2
        elif token_type == self.lexer.OROP:
            # disjunction
            phi_1 = self.visit(ctx.getRuleContext().getChild(0))
            phi_2 = self.visit(ctx.getRuleContext().getChild(2))

            return phi_1 + r" \/ " + phi_2

    # Visit a parse tree produced by stlParser#opReleaseExpr.
    def visitOpReleaseExpr(self, ctx: stlParser.OpReleaseExprContext) -> str:
        child_count = ctx.getRuleContext().getChildCount()

        phi_1 = ""
        phi_2 = ""

        if child_count == 4:
            index = self.freeze
            self.freeze += 1

            phi_1 = self.visit(ctx.getRuleContext().getChild(0))
            phi_2 = self.visit(ctx.getRuleContext().getChild(3))

            bounds = self.visit(ctx.getRuleContext().getChild(2))
            token_type = ctx.getRuleContext().getChild(1).getSymbol().type

            return self.freezeTime(index, bounds[0], bounds[1], [phi_1, phi_2], token_type)
        else:
            phi_1 = self.visit(ctx.getRuleContext().getChild(0))
            phi_2 = self.visit(ctx.getRuleContext().getChild(2))

            return phi_1 + " R " + phi_2

    # Visit a parse tree produced by stlParser#opNextExpr.
    def visitOpNextExpr(self, ctx: stlParser.OpNextExprContext) -> str:
        child_count = ctx.getRuleContext().getChildCount()

        phi = ""
        if child_count == 3:
            index = self.freeze
            self.freeze += 1

            bounds = self.visit(ctx.getRuleContext().getChild(1))
            phi = self.visit(ctx.getRuleContext().getChild(2))
            token_type = ctx.getRuleContext().getChild(0).getSymbol().type

            return self.freezeTime(index, bounds[0], bounds[1], [phi], token_type)
        else:
            phi = self.visit(ctx.getRuleContext().getChild(1))
            return "X " + phi

    # Visit a parse tree produced by stlParser#opPropExpr.
    def visitOpPropExpr(self, ctx: stlParser.OpPropExprContext) -> str:
        token_type = ctx.getRuleContext().getChild(1).getSymbol().type

        phi_1 = ""
        phi_2 = ""

        if token_type == self.lexer.IMPLIESOP:
            # implication
            phi_1 = self.visit(ctx.getRuleContext().getChild(0))
            phi_2 = self.visit(ctx.getRuleContext().getChild(2))

            return phi_1 + " -> " + phi_2
        elif token_type == self.lexer.EQUIVOP:
            # equivalency
            phi_1 = self.visit(ctx.getRuleContext().getChild(0))
            phi_2 = self.visit(ctx.getRuleContext().getChild(2))

            return phi_1 + " <-> " + phi_2

    # Visit a parse tree produced by stlParser#opNegExpr.
    def visitOpNegExpr(self, ctx: stlParser.OpNegExprContext) -> str:
        return "! " + self.visit(ctx.getRuleContext().getChild(1))

    # Visit a parse tree produced by stlParser#predicate.
    def visitPredicate(self, ctx: stlParser.PredicateContext) -> str:
        return ctx.getRuleContext().getChild(0).getText()

    # Visit a parse tree produced by stlParser#interval.
    def visitInterval(self, ctx: stlParser.IntervalContext) -> str:
        bounds = []
        bounds.append(float(ctx.getRuleContext().getChild(1).getText()))
        bounds.append(float(ctx.getRuleContext().getChild(3).getText()))

        return bounds


del stlParser



from .interval import Interval
from .model import BasicResult, ExtraResult, Model, ModelError, ModelInputs, ModelResult, Trace
from .optimizer import ObjectiveFn, Optimizer
from .result import Result, best_eval, best_run, worst_eval, worst_run
from .sample import Sample
from .scenario import Scenario
from .signal import Signal, SignalFactory
from .specification import Specification, SpecificationError

__all__ = [
    "Interval",
    "Model",
    "ModelInputs",
    "ModelResult",
    "BasicResult",
    "ExtraResult",
    "ModelError",
    "Optimizer",
    "ObjectiveFn",
    "Result",
    "Sample",
    "Scenario",
    "Signal",
    "SignalFactory",
    "Specification",
    "SpecificationError",
    "Trace",
    "best_eval",
    "best_run",
    "worst_eval",
    "worst_run",
]



from typing import Any, Callable, Dict, Iterable, Mapping, Sequence, Tuple

from attr import field, frozen

from .model import ModelInputs
from .sample import Sample
from .signal import Signal

SignalConstructor = Callable[[Sequence[float]], Signal]
ConstructorMap = Dict[Tuple[int, int], SignalConstructor]


def _to_static_params(value: Sequence[int]) -> slice:
    return slice(value[0], value[1], 1)


def _to_signals(mapping: Mapping[Tuple[int, int], SignalConstructor]) -> ConstructorMap:
    return dict(mapping)


def _validate_signals(_: Any, attr: Any, signals: ConstructorMap) -> None:
    for (lo, hi) in signals.keys():
        if (hi - lo) < 1:
            raise ValueError("Interval must contain at least one element")


def _signals_from_sample(ctors: ConstructorMap, sample: Sample) -> Iterable[Signal]:
    for (start, end), ctor in ctors.items():
        slice_len = end - start
        sample_range = slice(start, end, 1)
        signal_values = sample[sample_range]

        if len(signal_values) != slice_len:
            raise RuntimeError()

        yield ctor(sample[sample_range])


@frozen(slots=True)
class SampleLayout:
    """Representation of the layout of a sample."""

    static_parameters: slice = field(converter=_to_static_params)
    signals: ConstructorMap = field(converter=_to_signals, validator=_validate_signals)

    def decompose_sample(self, sample: Sample) -> ModelInputs:
        static_parameters = sample[self.static_parameters]
        signals = list(_signals_from_sample(self.signals, sample))

        return ModelInputs(static_parameters, signals)



from __future__ import annotations

from typing import Any, Iterable, List, Sequence, TypeVar, cast

import numpy as np
from attr import frozen

from .core.cost import SpecificationOrFactory
from .core.interval import Interval
from .core.layout import SampleLayout
from .core.model import Model, ModelResult
from .core.optimizer import Optimizer
from .core.result import Result
from .core.sample import Sample
from .core.scenario import Scenario
from .core.signal import Signal, SignalFactory
from .options import Options, SignalOptions

StateT = TypeVar("StateT")
CostT = TypeVar("CostT")
ResultT = TypeVar("ResultT")
ExtraT = TypeVar("ExtraT")


class SignalFactoryWrapper:
    def __init__(self, factory: SignalFactory, signal_times: List[float]):
        self.factory = factory
        self.signal_times = signal_times

    def __call__(self, signal_values: Sequence[float]) -> Signal:
        if len(signal_values) != len(self.signal_times):
            raise RuntimeError("not enough values for signal")

        signal = self.factory(self.signal_times, signal_values)

        if not isinstance(signal, Signal):
            raise ValueError("signal factory did not return signal")

        return signal


@frozen(slots=True)
class LayoutParameters:
    static_inputs: Sequence[Any]
    signal_options: Sequence[SignalOptions]
    t_span: Interval


def _create_sample_layout(params: LayoutParameters) -> SampleLayout:
    offset = len(params.static_inputs)
    static_inputs_range = (0, offset)
    signal_ranges_map = {}

    for signal_opts in params.signal_options:
        n_points = len(signal_opts.control_points)
        signal_range = (offset, offset + n_points)
        offset = offset + len(signal_opts.control_points)
        signal_times = signal_opts.signal_times

        if signal_times is None:
            times_array = np.linspace(
                params.t_span.lower,
                params.t_span.upper,
                num=n_points,
                dtype=np.float64,
            )
            signal_times = cast(List[float], times_array.tolist())

        signal_ranges_map[signal_range] = SignalFactoryWrapper(signal_opts.factory, signal_times)

    return SampleLayout(static_inputs_range, signal_ranges_map)


def _signal_bounds(signals: Iterable[SignalOptions]) -> tuple[Interval, ...]:
    return sum((signal.control_points for signal in signals), ())


def staliro(
    model: Model[StateT, ExtraT],
    specification: SpecificationOrFactory[StateT, CostT],
    optimizer: Optimizer[CostT, ResultT],
    options: Options,
) -> Result[ResultT, CostT, ExtraT]:
    """Search for falsifying inputs to the provided system.

    Using the optimizer, search the input space defined in the options for cases which falsify the
    provided specification represented by the formula phi and the predicates.

    Args:
        model: The model of the system being tested.
        specification: The requirement for the system.
        optimizer: The optimizer to use to search the sample space.
        options: General options to manipulate the behavior of the search.

    Returns:
        An object containing the result of each optimization attempt.
    """
    params = LayoutParameters(options.static_parameters, options.signals, options.interval)
    layout = _create_sample_layout(params)
    signal_bounds = _signal_bounds(options.signals)
    bounds = options.static_parameters + signal_bounds

    scenario: Scenario[StateT, CostT, ResultT, ExtraT] = Scenario(
        model=model,
        specification=specification,
        runs=options.runs,
        iterations=options.iterations,
        seed=options.seed,
        processes=options.process_count,
        bounds=bounds,
        interval=options.interval,
        layout=layout,
    )

    return scenario.run(optimizer)


def simulate_model(
    model: Model[StateT, ExtraT], options: Options, sample: Sample
) -> ModelResult[StateT, ExtraT]:
    params = LayoutParameters(options.static_parameters, options.signals, options.interval)
    layout = _create_sample_layout(params)
    inputs = layout.decompose_sample(sample)

    return model.simulate(inputs, options.interval)



from __future__ import annotations

import math
from typing import Callable, List, Optional, Sequence, TypeVar, Union, cast

import numpy as np
from numpy.typing import NDArray
from scipy import integrate
from typing_extensions import Literal, overload

from .core.interval import Interval
from .core.model import BasicResult, Model, ModelInputs, ModelResult, Trace
from .core.signal import Signal

StateT = TypeVar("StateT")
ExtraT = TypeVar("ExtraT")

Signals = Sequence[Signal]
StaticParameters = Sequence[float]
SignalTimes = NDArray[np.float_]
SignalValues = NDArray[np.float_]
BlackboxFunc = Callable[[StaticParameters, SignalTimes, SignalValues], ModelResult[StateT, ExtraT]]


class Blackbox(Model[StateT, ExtraT]):
    """General system model which does not make assumptions about the underlying system.

    Attributes:
        func: User-defined function which is given the static parameters, a vector of time values,
              and a matrix of interpolated signal values and returns a SimulationResult or
              Falsification
        sampling_interval: Time-step to use for interpolating signal values over the simulation
                           interval
    """

    def __init__(self, func: BlackboxFunc[StateT, ExtraT], sampling_interval: float):
        self.func = func
        self.sampling_interval = sampling_interval

    def simulate(self, inputs: ModelInputs, interval: Interval) -> ModelResult[StateT, ExtraT]:
        step_count = math.floor(interval.length / self.sampling_interval)
        signal_times = np.linspace(start=interval.lower, stop=interval.upper, num=step_count)
        signal_times_list: List[float] = signal_times.tolist()
        signal_traces = [signal.at_times(signal_times_list) for signal in inputs.signals]

        return self.func(inputs.static, signal_times, np.array(signal_traces))


State = NDArray[np.float_]
IntegrationFn = Callable[[float, State], State]
ODEFunc = Callable[[float, State, SignalValues], State]
ODEResult = ModelResult[List[float], None]


class IntegrationFunc:
    def __init__(self, signals: Sequence[Signal], ode_func: ODEFunc):
        self.signals = signals
        self.func = ode_func

    def __call__(self, time: float, state: State) -> State:
        signal_values = np.array([signal.at_time(time) for signal in self.signals])
        return self.func(time, state, signal_values)


_MethodT = Literal["RK45", "RK23", "DOP853", "Radau", "BDF", "LSODA"]


class ODE(Model[List[float], None]):
    """Model which assumes the underlying system is modeled by an Ordinary Differential Equation.

    Attributes:
        func: User-defined function which is given a time value, a state, and a vector of signal
              values and returns a new state.
    """

    def __init__(self, func: ODEFunc, method: _MethodT):
        self.func = func
        self.method = method

    def simulate(self, inputs: ModelInputs, interval: Interval) -> ODEResult:
        integration_fn = IntegrationFunc(inputs.signals, self.func)
        integration = integrate.solve_ivp(
            fun=integration_fn,
            t_span=interval.astuple(),
            y0=inputs.static,
            method=self.method,
        )
        times = cast(List[float], integration.t.tolist())
        states = cast(List[List[float]], integration.y.tolist())
        trace = Trace(times, states)

        return BasicResult(trace)


_BlackboxDecorator = Callable[[BlackboxFunc[StateT, ExtraT]], Blackbox[StateT, ExtraT]]


@overload
def blackbox(
    *, sampling_interval: float = ...
) -> Callable[[BlackboxFunc[StateT, ExtraT]], Blackbox[StateT, ExtraT]]:
    ...


@overload
def blackbox(_func: BlackboxFunc[StateT, ExtraT]) -> Blackbox[StateT, ExtraT]:
    ...


def blackbox(
    _func: Optional[BlackboxFunc[StateT, ExtraT]] = None, *, sampling_interval: float = 0.1
) -> Union[Blackbox[StateT, ExtraT], _BlackboxDecorator[StateT, ExtraT]]:
    """Decorate a function as a blackbox model.

    This decorator can be used with or without arguments.

    Args:
        func: The function to wrap as a blackbox
        *: Variable length argument list
        sampling_interval: The time interval to use when sampling the signal interpolators

    Returns:
        A blackbox model
    """

    def decorator(func: BlackboxFunc[StateT, ExtraT]) -> Blackbox[StateT, ExtraT]:
        return Blackbox(func, sampling_interval)

    if _func is not None:
        return decorator(_func)
    else:
        return decorator


_ODEDecorator = Callable[[ODEFunc], ODE]


@overload
def ode(*, method: _MethodT = ...) -> _ODEDecorator:
    ...


@overload
def ode(_func: ODEFunc) -> ODE:
    ...


def ode(_func: Optional[ODEFunc] = None, *, method: _MethodT = "RK45") -> Union[_ODEDecorator, ODE]:
    """Decorate a function as an ODE model.

    This decorator can be used with or without arguments.

    Args:
        _func: The function to wrap as an ODE model

    Returns:
        An ODE model
    """

    def decorator(func: ODEFunc) -> ODE:
        return ODE(func, method)

    if _func is not None:
        return decorator(_func)
    else:
        return decorator



from __future__ import annotations

import logging
import time
from concurrent.futures import ProcessPoolExecutor
from typing import Callable, Generic, Iterable, Iterator, Sequence, Tuple, TypeVar, Union

from attr import field, frozen
from attr.validators import instance_of

from .interval import Interval
from .layout import SampleLayout
from .model import FailureResult, Model, ModelResult
from .optimizer import ObjectiveFn
from .result import Evaluation, TimingData
from .sample import Sample
from .specification import Specification, SpecificationError

StateT = TypeVar("StateT")
CostT = TypeVar("CostT")
SpecificationFactory = Callable[[Sample], Specification[StateT, CostT]]
SpecificationOrFactory = Union[Specification[StateT, CostT], SpecificationFactory[StateT, CostT]]

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class EvaluationError(Exception):
    pass


T = TypeVar("T")


def _time(func: Callable[[], T]) -> Tuple[float, T]:
    start_time = time.perf_counter()
    result = func()
    stop_time = time.perf_counter()
    duration = stop_time - start_time

    return duration, result


ExtraT = TypeVar("ExtraT")


@frozen()
class Thunk(Generic[StateT, CostT, ExtraT]):
    """Class which represents the deferred evaluation of the cost function.

    A Thunk contains all the necessary information to produce an Evaluation, but does not execute
    the computation until the evaluate method is called (lazy value).

    Attributes:
        sample: The sample generated by the optimizer that is provided to the model
        model: Model instance which represents the system
        specification: Specification instance which represents the requirement
        options: Configuration object that controls the behavior of the Model and Specification
    """

    sample: Sample = field(validator=instance_of(Sample))
    model: Model[StateT, ExtraT]
    _specification: SpecificationOrFactory[StateT, CostT]
    interval: Interval
    layout: SampleLayout

    @property
    def specification(self) -> Specification[StateT, CostT]:
        if not callable(self._specification):
            return self._specification

        specification = self._specification(self.sample)

        if not isinstance(specification, Specification):
            raise SpecificationError(f"invalid type {type(specification)} given as specification")

        return specification

    def evaluate(self) -> Evaluation[CostT, ExtraT]:
        """Evaluate the sample using the specification and model.

        The computation is the following pipeline:
            sample -> model.simulate(sample) -> specification.evaluate(model_result) -> cost

        Returns:
            An Evaluation instance representing the result of the computation pipeline.
        """

        inputs = self.layout.decompose_sample(self.sample)
        simulate = lambda: self.model.simulate(inputs, self.interval)
        model_duration, model_result = _time(simulate)

        if not isinstance(model_result, ModelResult):
            raise EvaluationError(f"Incorrect return type from model {type(model_result)}")

        if isinstance(model_result, FailureResult):
            cost = self.specification.failure_cost
            cost_duration = 0.0
        else:
            trace = model_result.trace
            compute_cost = lambda: self.specification.evaluate(trace.states, trace.times)
            cost_duration, cost = _time(compute_cost)

        timing_data = TimingData(model_duration, cost_duration)

        return Evaluation(cost, self.sample, model_result.extra, timing_data)


@frozen()
class ThunkGenerator(Generic[StateT, CostT, ExtraT], Iterable[Thunk[StateT, CostT, ExtraT]]):
    """Generate Thunk instances from a collection of samples."""

    samples: Iterable[Sample]
    model: Model[StateT, ExtraT]
    specification: SpecificationOrFactory[StateT, CostT]
    interval: Interval
    layout: SampleLayout

    def __iter__(self) -> Iterator[Thunk[StateT, CostT, ExtraT]]:
        for sample in self.samples:
            yield Thunk(
                sample,
                self.model,
                self.specification,
                self.interval,
                self.layout,
            )


@frozen()
class CostFn(Generic[StateT, CostT, ExtraT], ObjectiveFn[CostT]):
    """Class which represents the composition of a Model and Specification.

    A Model is responsible for modeling the system and returning a trajectory given a sample, and a
    Specification is responsible for evaluating a system trajectory and returning a value which
    indicates its "goodness". Composing these two components produces a component which can produce
    "goodness" values given a sample.

    Attributes:
        model: Model instance which represents a system
        specification: Specification instance or factory which represents the requirement
        options: Configuration object that controls the behavior of the Model and Specification
        history: List of all evaluations produced by this instance
    """

    model: Model[StateT, ExtraT]
    specification: SpecificationOrFactory[StateT, CostT]
    interval: Interval
    layout: SampleLayout
    history: list[Evaluation[CostT, ExtraT]] = field(init=False, factory=list)

    def eval_sample(self, sample: Sample) -> CostT:
        """Compute the cost of a single sample.

        Args:
            sample: Sample to evaluate

        Returns:
            The sample cost
        """

        logger.debug(f"Evaluating sample {sample}")

        thunk = Thunk(
            sample,
            self.model,
            self.specification,
            self.interval,
            self.layout,
        )
        evaluation = thunk.evaluate()

        self.history.append(evaluation)

        return evaluation.cost

    def eval_samples(self, samples: Sequence[Sample]) -> list[CostT]:
        """Compute the cost of multiple samples sequentially.

        Args:
            samples: Samples to evaluate

        Returns:
            The cost for each sample
        """

        logger.debug(f"Evaluating samples {samples}")

        thunks = ThunkGenerator(
            samples,
            self.model,
            self.specification,
            self.interval,
            self.layout,
        )
        evaluations = [thunk.evaluate() for thunk in thunks]

        self.history.extend(evaluations)

        return [evaluation.cost for evaluation in evaluations]

    def eval_samples_parallel(self, samples: Sequence[Sample], processes: int) -> list[CostT]:
        """Compute the cost of multiple samples in parallel.

        Samples are evaluated row-wise, so each row is considered a different sample.

        Args:
            samples: The samples to evaluate
            processes: Number of processes to use to evaluate the samples

        Returns:
            The cost for each sample.
        """

        logger.debug(f"Evaluating samples {samples} with {processes} processes")

        thunks = ThunkGenerator(
            samples,
            self.model,
            self.specification,
            self.interval,
            self.layout,
        )

        with ProcessPoolExecutor(max_workers=processes) as executor:
            futures: Iterable[Evaluation[CostT, ExtraT]] = executor.map(Thunk.evaluate, thunks)
            evaluations = list(futures)

        self.history.extend(evaluations)

        return [evaluation.cost for evaluation in evaluations]



from __future__ import annotations

from typing import Sequence, TypeVar

from typing_extensions import Protocol, runtime_checkable

StateT = TypeVar("StateT", contravariant=True)
CostT = TypeVar("CostT", covariant=True)


@runtime_checkable
class Specification(Protocol[StateT, CostT]):
    """Class that represents a requirement to be evaluated using simulation data.

    A specification should accept the trajectories and timestamps generated by a model and return
    a value which represents the "goodness" of the model results with respect to some criteria.
    """

    @property
    def failure_cost(self) -> CostT:
        ...

    def evaluate(self, state: Sequence[StateT], timestamps: Sequence[float]) -> CostT:
        """Evaluate trajectories and timestamps with respect to some requirement.

        Args:
            trajectories: Matrix of states where each row represents a state variable.
            timestamps: Vector of time values corresponding to each column of the trajectories
                        matrix.

        Returns:
            A value indicating the "goodness" of the trajectories according to the requirement.
        """
        ...


class SpecificationError(Exception):
    pass



from __future__ import annotations

from typing import Any, Iterable, Iterator, Sequence, Sized, Tuple, Union, cast

from attr import field, frozen
from numpy import float64, ndarray
from numpy.typing import NDArray
from typing_extensions import overload


def _value_converter(value: Any) -> float:
    if isinstance(value, int):
        return float(value)

    if isinstance(value, float):
        return value

    raise TypeError("only int and float are valid sample value types")


_ValuesT = Union[NDArray[Any], Sequence[int], Sequence[float]]


def _values_converter(value: _ValuesT) -> Tuple[float, ...]:
    if isinstance(value, ndarray):
        return cast(Tuple[float, ...], tuple(value.astype(float64).tolist()))

    if isinstance(value, (list, tuple)):
        return tuple(_value_converter(v) for v in value)

    raise TypeError("only ndarray, list and tuple are valid collections to create a sample")


@frozen(slots=True)
class Sample(Sized, Iterable[float]):
    """Representation of a sample generated by the optimizer.

    Args:
        values: The sample values

    Attributes:
        values: The sample values
    """

    values: Tuple[float, ...] = field(converter=_values_converter)

    def __len__(self) -> int:
        return len(self.values)

    def __iter__(self) -> Iterator[float]:
        return iter(self.values)

    @overload
    def __getitem__(self, index: int) -> float:
        ...

    @overload
    def __getitem__(self, index: slice) -> Tuple[float, ...]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[float, Tuple[float, ...]]:
        if isinstance(index, (int, slice)):
            return self.values[index]
        else:
            raise TypeError()



from __future__ import annotations

from typing import Iterable, Iterator, Tuple, Union

from attr import Attribute, field, frozen

BoundT = Union[int, float]


class IntervalError(Exception):
    pass


def _bound_converter(bound: BoundT) -> float:
    if isinstance(bound, int):
        return float(bound)

    if isinstance(bound, float):
        return bound

    raise IntervalError(f"Expected [float, int] type for bound, recieved: {type(bound)}")


def _upper_validator(inst: Interval, _: Attribute[float], upper: float) -> None:
    if upper == inst.lower:
        raise IntervalError("interval cannot have zero length")

    if upper < inst.lower:
        raise IntervalError("interval upper bound must be greater than lower bound")


@frozen(slots=True)
class Interval(Iterable[float]):
    """Representation of an interval of values.

    Args:
        lower: The lower bound of the interval
        upper: The upper bound of the interval

    Attributes:
        bounds: The upper and lower values of the interval
        length: The length of the interval
    """

    lower: float = field(converter=_bound_converter)
    upper: float = field(converter=_bound_converter, validator=_upper_validator)

    def __iter__(self) -> Iterator[float]:
        return iter(self.astuple())

    @property
    def length(self) -> float:
        """The length of the interval."""

        return self.upper - self.lower

    def astuple(self) -> Tuple[float, float]:
        return (self.lower, self.upper)



# Generated from stlParser.g4 by ANTLR 4.5.1
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .stlParser import stlParser
else:
    from stlParser import stlParser

import tltk_mtl as mtl
import numpy as np

# This class defines a complete generic visitor for a parse tree produced by stlParser.
class stlParserVisitor(ParseTreeVisitor):
    def __init__(self, lexer, predicates, mode):
        if not isinstance(predicates, (dict, list, tuple)):
            raise ValueError("predicates must be list, dict or tuple")

        if isinstance(predicates, dict):
            if not all(isinstance(pred, mtl.Predicate) for pred in predicates.values()):
                raise ValueError("all dictionary values must be TLTK Predicate objects")

        self._lexer = lexer
        self._mode = mode
        self._predicates = predicates

    # Visit a parse tree produced by stlParser#stlSpecification.
    def visitStlSpecification(self, ctx: stlParser.StlSpecificationContext):
        return self.visit(ctx.getRuleContext().getChild(0))

    # Visit a parse tree produced by stlParser#predicateExpr.
    def visitPredicateExpr(self, ctx: stlParser.PredicateExprContext):
        return self.visit(ctx.getRuleContext().getChild(0))

    # Visit a parse tree produced by stlParser#opFutureExpr.
    def visitOpFutureExpr(self, ctx: stlParser.OpFutureExprContext):
        pred = self.visit(ctx.getRuleContext().getChild(2))
        bounds = self.visit(ctx.getRuleContext().getChild(1))

        return mtl.Finally(bounds[0], bounds[1], pred, self._mode)

    # Visit a parse tree produced by stlParser#parenPhiExpr.
    def visitParenPhiExpr(self, ctx: stlParser.ParenPhiExprContext):
        return self.visit(ctx.getRuleContext().getChild(1))

    # Visit a parse tree produced by stlParser#opUntilExpr.
    def visitOpUntilExpr(self, ctx: stlParser.OpUntilExprContext):
        pred1 = self.visit(ctx.getRuleContext().getChild(0))
        pred2 = self.visit(ctx.getRuleContext().getChild(3))

        bounds = self.visit(ctx.getRuleContext().getChild(2))

        return mtl.Until(bounds[0], bounds[1], pred1, pred2, self._mode)

    # Visit a parse tree produced by stlParser#opGloballyExpr.
    def visitOpGloballyExpr(self, ctx: stlParser.OpGloballyExprContext):
        pred = self.visit(ctx.getRuleContext().getChild(2))
        bounds = self.visit(ctx.getRuleContext().getChild(1))

        return mtl.Global(bounds[0], bounds[1], pred, self._mode)

    # Visit a parse tree produced by stlParser#opLogicalExpr.
    def visitOpLogicalExpr(self, ctx: stlParser.OpLogicalExprContext):
        pred1 = self.visit(ctx.getRuleContext().getChild(0))
        pred2 = self.visit(ctx.getRuleContext().getChild(2))

        type = ctx.getRuleContext().getChild(1).getSymbol().type

        if type == self._lexer.ANDOP:
            return mtl.And(pred1, pred2, self._mode)
        elif type == self._lexer.OROP:
            return mtl.Or(pred1, pred2, self._mode)

    # Visit a parse tree produced by stlParser#opReleaseExpr.
    def visitOpReleaseExpr(self, ctx: stlParser.OpReleaseExprContext):
        pred1 = self.visit(ctx.getRuleContext().getChild(0))
        pred2 = self.visit(ctx.getRuleContext().getChild(3))

        bounds = self.visit(ctx.getRuleContext().getChild(2))

        return mtl.Not(mtl.Until(bounds[0], bounds[1], mtl.Not(pred1), mtl.Not(pred2)))

    # Visit a parse tree produced by stlParser#opNextExpr.
    def visitOpNextExpr(self, ctx: stlParser.OpNextExprContext):
        pred = self.visit(ctx.getRuleContext().getChild(1))

        return mtl.Next(pred)

    # Visit a parse tree produced by stlParser#opPropExpr.
    def visitOpPropExpr(self, ctx: stlParser.OpPropExprContext):
        pred1 = self.visit(ctx.getRuleContext().getChild(0))
        pred2 = self.visit(ctx.getRuleContext().getChild(2))

        return mtl.Or(mtl.Not(pred1), pred2, self._mode)

    # Visit a parse tree produced by stlParser#opNegExpr.
    def visitOpNegExpr(self, ctx: stlParser.OpNegExprContext):
        pred = self.visit(ctx.getRuleContext().getChild(1))

        return mtl.Not(pred, self._mode)

    # Visit a parse tree produced by stlParser#predicate.
    def visitPredicate(self, ctx: stlParser.PredicateContext):
        if ctx.getRuleContext().getChildCount() == 1:
            child_name = ctx.getRuleContext().getChild(0).getText()

            # the predicate name only exists, so return the relevant
            # mtl.Predicate data structure
            if not isinstance(self._predicates, dict):
                raise ValueError(
                    "singular variable names require predicates be provided as dictionary"
                )

            return self._predicates[child_name]
        else:
            minus: str = ctx.getRuleContext().getChild(0).getText()

            if minus == "-":
                var: str = ctx.getRuleContext().getChild(1).getText()
                operator: str = ctx.getRuleContext().getChild(2).getText()
                value: str = ctx.getRuleContext().getChild(3).getText()
            else:
                var: str = ctx.getRuleContext().getChild(0).getText()
                operator: str = ctx.getRuleContext().getChild(1).getText()
                value: str = ctx.getRuleContext().getChild(2).getText()

            # check that the variable is valid
            if var not in self._predicates:
                raise Exception(
                    f"Error: predicate {var} is not in the list of valid variables {self._predicates}"
                )

            if operator == "<" or operator == ">":
                raise Exception(
                    "Warning: strict relational operator used. Please use non-strict (<= or >=)"
                )
            elif operator == "<=":
                if minus == "-":
                    return mtl.Predicate(var, -1, float(value))
                else:
                    return mtl.Predicate(var, 1, float(value))
            elif operator == ">=":
                if minus == "-":
                    return mtl.Predicate(var, 1, -float(value))
                else:
                    return mtl.Predicate(var, -1, -float(value))

    # Visit a parse tree produced by stlParser#interval.
    def visitInterval(self, ctx: stlParser.IntervalContext):
        bounds = []
        bounds.append(float(ctx.getRuleContext().getChild(1).getText()))
        bounds.append(float(ctx.getRuleContext().getChild(3).getText()))

        return bounds



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Sequence, TypeVar

from attr import frozen

from .interval import Interval
from .signal import Signal

StateT = TypeVar("StateT")
ExtraT = TypeVar("ExtraT")


@frozen(eq=False)
class Trace(Generic[StateT]):
    _times: Sequence[float]
    _states: Sequence[StateT]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Trace):
            return NotImplemented

        return self.times == other.times and self.states == other.states

    def __len__(self) -> int:
        return len(self._times)

    @property
    def times(self) -> list[float]:
        return list(self._times)

    @property
    def states(self) -> list[StateT]:
        return list(self._states)


class ModelResult(Generic[StateT, ExtraT], ABC):
    @property
    @abstractmethod
    def trace(self) -> Trace[StateT]:
        ...

    @property
    @abstractmethod
    def extra(self) -> ExtraT:
        ...


class BasicResult(ModelResult[StateT, None]):
    def __init__(self, trace: Trace[StateT]):
        self._trace = trace

    @property
    def trace(self) -> Trace[StateT]:
        return self._trace

    @property
    def extra(self) -> None:
        return None


class ExtraResult(ModelResult[StateT, ExtraT]):
    def __init__(self, trace: Trace[StateT], extra: ExtraT):
        self._trace = trace
        self._extra = extra

    @property
    def trace(self) -> Trace[StateT]:
        return self._trace

    @property
    def extra(self) -> ExtraT:
        return self._extra


class FailureResult(ModelResult[StateT, ExtraT]):
    def __init__(self, extra: ExtraT):
        self._extra = extra

    @property
    def trace(self) -> Trace[StateT]:
        return Trace([], [])

    @property
    def extra(self) -> ExtraT:
        return self._extra


@frozen()
class ModelInputs:
    static: Sequence[float]
    signals: Sequence[Signal]


class Model(Generic[StateT, ExtraT], ABC):
    """A model is a representation of a system under test (SUT).

    The model defines a simulate method that contains the logic necessary to simulate the system
    given a set of inputs.
    """

    @abstractmethod
    def simulate(self, inputs: ModelInputs, interval: Interval) -> ModelResult[StateT, ExtraT]:
        """Simulate the model using the given inputs.

        This method contains the logic responsible for simulating the model with the given inputs.
        The result of this method should be a ModelResult instance containing a set of timestamps
        and state values representing the execution of the system over time, or a Failure instance
        that serves as an indicator of a system failure that should be interpreted as a
        falsification of the system requirement.

        Arguments:
            static: The static inputs to the system
            signals: time-varying inputs to the system represented as interpolated functions

        Returns:
            An instance of ModelData indicating a successful simulation, or a Failure if an error is
            encountered that should result in a falsification.
        """
        ...


class ModelError(Exception):
    pass



from __future__ import annotations

import os
import random
from typing import Any, List, Optional, Sequence, Tuple, Union, cast

import numpy as np
from attr import Attribute, converters, field, frozen, validators
from numpy.typing import NDArray
from typing_extensions import Literal

from .core.interval import BoundT, Interval
from .core.signal import SignalFactory
from .signals import pchip


class OptionsError(Exception):
    pass


_IntervalT = Union[Sequence[BoundT], NDArray[Any]]


def _to_interval(interval: Union[Interval, _IntervalT]) -> Interval:
    """Convert a value to an interval.

    This function only supports ordered collections because the order of values in the iterable
    matters when trying to construct an interval. Iterables like Set do not guarantee the order of
    iteration, which can lead to non-deterministic failures.

    Arguments:
        value: The value to convert to an interval. Must be an Interval instance, list or tuple

    Returns:
        An instance of Interval using the values provided in the ordered collection
    """

    if isinstance(interval, Interval):
        return interval

    if isinstance(interval, np.ndarray):
        interval = interval.astype(float)

    return Interval(interval[0], interval[1])


_IntervalsT = Sequence[_IntervalT]


def _to_intervals(intervals: _IntervalsT) -> Tuple[Interval, ...]:
    """Convert a sequence of values into a sequence on intervals."""

    return tuple(_to_interval(interval) for interval in intervals)


def _to_signal_times(times: Union[Sequence[float], NDArray[np.float_]]) -> list[float]:
    if isinstance(times, np.ndarray):
        return cast(List[float], np.array(times, dtype=float).tolist())

    return list(times)


@frozen()
class SignalOptions:
    """Options for signal generation.

    Attributes:
        interval: The interval the signal should be generated for
        factory: Factory to produce interpolators for the signal
        control_points: The number of points the optimizer should generate for the signal
        signal_times: Time values for the generated control points
        time_varying: Flag that indicates that the signal times should be considered a search
                      variable (EXPERIMENTAL)
    """

    control_points: Tuple[Interval, ...] = field(converter=_to_intervals)
    factory: SignalFactory = field(default=pchip, validator=validators.is_callable())
    signal_times: Optional[list[float]] = field(
        default=None,
        converter=converters.optional(_to_signal_times),
        validator=validators.optional(
            validators.deep_iterable(validators.instance_of(float), validators.instance_of(list))
        ),
    )
    time_varying: bool = field(default=False)


def _seed_factory() -> int:
    return random.randint(0, 2**32 - 1)


_ParallelizationT = Union[None, Literal["all", "cores"], int]


def _parallelization_validator(_: Any, attr: Attribute[Any], value: _ParallelizationT) -> None:
    is_none = value is None
    is_valid_str = value in {"all", "cores"}
    is_int = isinstance(value, int)

    if not is_none and not is_valid_str and not is_int:
        raise OptionsError(
            "invalid parallelization value. Allowed types are: [None, 'all', 'cores', int]"
        )


@frozen()
class Options:
    """General options for controlling falsification behavior.

    Attributes:
        static_parameters: Parameters that will be provided to the system at the beginning and are
            time invariant (initial conditions)
        signals: System inputs that will vary over time
        seed: The initial seed of the random number generator
        iterations: The number of search iterations to perform in a run
        runs: The number times to run the optimizer
        interval: The time interval of the system simulation
        parallelization: Number of processes to use to parallelize runs of the optimizer. Acceptable
            values are: "cores" (all available cores), "all" (all runs), int (number of processes),
            None (no parallelization).
    """

    static_parameters: Tuple[Interval, ...] = field(factory=tuple, converter=_to_intervals)
    signals: Sequence[SignalOptions] = field(
        factory=list, validator=validators.deep_iterable(validators.instance_of(SignalOptions))
    )
    seed: int = field(
        factory=_seed_factory, validator=[validators.instance_of(int), validators.gt(0)]
    )
    iterations: int = field(default=400, validator=[validators.instance_of(int), validators.gt(0)])
    runs: int = field(default=1, validator=[validators.instance_of(int), validators.gt(0)])
    interval: Interval = field(default=Interval(0.0, 10.0), converter=_to_interval)
    parallelization: _ParallelizationT = field(default=None, validator=_parallelization_validator)

    @property
    def process_count(self) -> Optional[int]:
        """Number of processes to use based on the value of the parallelization attribute."""

        if self.parallelization == "all":
            return self.runs
        elif self.parallelization == "cores":
            return os.cpu_count()
        else:
            return self.parallelization



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Sequence, TypeVar

from typing_extensions import Protocol

from .interval import Interval
from .sample import Sample

CostT = TypeVar("CostT", covariant=True)


class ObjectiveFn(Protocol[CostT]):
    """Representation of a function that can be optimized by an optimizer.

    An objective function evaluates samples generated by the optimizer and returns a scalar cost
    value for each sample. The objective function is capable of evaluating one or more samples at
    once.
    """

    def eval_sample(self, sample: Sample) -> CostT:
        """Evaluate a single sample.

        Arguments:
            sample: The sample to evaluate

        Returns:
            The cost of the sample
        """
        ...

    def eval_samples(self, samples: Sequence[Sample]) -> Sequence[CostT]:
        """Evaluate several samples sequentially.

        Arguments:
            samples: A sequence of samples to evaluate

        Returns:
            The cost of each sample in the same order they were given
        """
        ...

    def eval_samples_parallel(self, samples: Sequence[Sample], processes: int) -> Sequence[CostT]:
        """Evaluate several samples parallelizing according to the number of processes.

        Arguments:
            samples: A sequence of samples to evaluate
            processes: The number of processes to use to evaluate the samples

        Returns:
            The cost of each sample in the sample order they were given
        """
        ...


ResultT = TypeVar("ResultT")


class Optimizer(Generic[CostT, ResultT], ABC):
    """An optimizer selects samples to be evaluated by the cost function."""

    @abstractmethod
    def optimize(
        self, func: ObjectiveFn[CostT], bounds: Sequence[Interval], budget: int, seed: int
    ) -> ResultT:
        """Evaluate samples and use the result to select more samples until the budget is reached.

        The optimize method is responsible for generating samples from the
        """
        ...



# Generated from stlParser.g4 by ANTLR 4.7.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\33")
        buf.write("M\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\3\2\3\2\3\2\3\3\3\3")
        buf.write("\3\3\3\3\3\3\3\3\3\3\3\3\3\3\5\3\27\n\3\3\3\3\3\3\3\5")
        buf.write("\3\34\n\3\3\3\3\3\3\3\5\3!\n\3\3\3\3\3\5\3%\n\3\3\3\3")
        buf.write("\3\3\3\5\3*\n\3\3\3\3\3\3\3\3\3\5\3\60\n\3\3\3\3\3\3\3")
        buf.write("\3\3\3\3\3\3\3\3\7\39\n\3\f\3\16\3<\13\3\3\4\5\4?\n\4")
        buf.write("\3\4\3\4\3\4\3\4\5\4E\n\4\3\5\3\5\3\5\3\5\3\5\3\5\3\5")
        buf.write("\2\3\4\6\2\4\6\b\2\7\3\2\22\23\3\2\24\25\4\2\4\4\6\6\4")
        buf.write("\2\26\26\30\30\4\2\5\5\7\7\2X\2\n\3\2\2\2\4$\3\2\2\2\6")
        buf.write("D\3\2\2\2\bF\3\2\2\2\n\13\5\4\3\2\13\f\7\2\2\3\f\3\3\2")
        buf.write("\2\2\r\16\b\3\1\2\16\17\7\4\2\2\17\20\5\4\3\2\20\21\7")
        buf.write("\5\2\2\21%\3\2\2\2\22\23\7\n\2\2\23%\5\4\3\13\24\26\7")
        buf.write("\r\2\2\25\27\5\b\5\2\26\25\3\2\2\2\26\27\3\2\2\2\27\30")
        buf.write("\3\2\2\2\30%\5\4\3\n\31\33\7\16\2\2\32\34\5\b\5\2\33\32")
        buf.write("\3\2\2\2\33\34\3\2\2\2\34\35\3\2\2\2\35%\5\4\3\t\36 \7")
        buf.write("\17\2\2\37!\5\b\5\2 \37\3\2\2\2 !\3\2\2\2!\"\3\2\2\2\"")
        buf.write("%\5\4\3\b#%\5\6\4\2$\r\3\2\2\2$\22\3\2\2\2$\24\3\2\2\2")
        buf.write("$\31\3\2\2\2$\36\3\2\2\2$#\3\2\2\2%:\3\2\2\2&\'\f\7\2")
        buf.write("\2\')\7\20\2\2(*\5\b\5\2)(\3\2\2\2)*\3\2\2\2*+\3\2\2\2")
        buf.write("+9\5\4\3\b,-\f\6\2\2-/\7\21\2\2.\60\5\b\5\2/.\3\2\2\2")
        buf.write("/\60\3\2\2\2\60\61\3\2\2\2\619\5\4\3\7\62\63\f\5\2\2\63")
        buf.write("\64\t\2\2\2\649\5\4\3\6\65\66\f\4\2\2\66\67\t\3\2\2\67")
        buf.write("9\5\4\3\58&\3\2\2\28,\3\2\2\28\62\3\2\2\28\65\3\2\2\2")
        buf.write("9<\3\2\2\2:8\3\2\2\2:;\3\2\2\2;\5\3\2\2\2<:\3\2\2\2=?")
        buf.write("\7\t\2\2>=\3\2\2\2>?\3\2\2\2?@\3\2\2\2@A\7\27\2\2AB\7")
        buf.write("\13\2\2BE\7\30\2\2CE\7\27\2\2D>\3\2\2\2DC\3\2\2\2E\7\3")
        buf.write("\2\2\2FG\t\4\2\2GH\t\5\2\2HI\7\b\2\2IJ\t\5\2\2JK\t\6\2")
        buf.write("\2K\t\3\2\2\2\f\26\33 $)/8:>D")
        return buf.getvalue()


class stlParser ( Parser ):

    grammarFileName = "stlParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "<INVALID>", "'('", "')'", "'['", "']'", 
                     "','", "'-'", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "'inf'" ]

    symbolicNames = [ "<INVALID>", "WS", "LPAREN", "RPAREN", "LBRACK", "RBRACK", 
                      "COMMA", "MINUS", "NEGATION", "RELOP", "EQUALITYOP", 
                      "NEXTOP", "FUTUREOP", "GLOBALLYOP", "UNTILOP", "RELEASEOP", 
                      "ANDOP", "OROP", "IMPLIESOP", "EQUIVOP", "INF", "NAME", 
                      "NUMBER", "INT_NUMBER", "FLOAT_NUMBER", "SCIENTIFIC_NUMBER" ]

    RULE_stlSpecification = 0
    RULE_phi = 1
    RULE_predicate = 2
    RULE_interval = 3

    ruleNames =  [ "stlSpecification", "phi", "predicate", "interval" ]

    EOF = Token.EOF
    WS=1
    LPAREN=2
    RPAREN=3
    LBRACK=4
    RBRACK=5
    COMMA=6
    MINUS=7
    NEGATION=8
    RELOP=9
    EQUALITYOP=10
    NEXTOP=11
    FUTUREOP=12
    GLOBALLYOP=13
    UNTILOP=14
    RELEASEOP=15
    ANDOP=16
    OROP=17
    IMPLIESOP=18
    EQUIVOP=19
    INF=20
    NAME=21
    NUMBER=22
    INT_NUMBER=23
    FLOAT_NUMBER=24
    SCIENTIFIC_NUMBER=25

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class StlSpecificationContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def phi(self):
            return self.getTypedRuleContext(stlParser.PhiContext,0)


        def EOF(self):
            return self.getToken(stlParser.EOF, 0)

        def getRuleIndex(self):
            return stlParser.RULE_stlSpecification

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStlSpecification" ):
                listener.enterStlSpecification(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStlSpecification" ):
                listener.exitStlSpecification(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitStlSpecification" ):
                return visitor.visitStlSpecification(self)
            else:
                return visitor.visitChildren(self)




    def stlSpecification(self):

        localctx = stlParser.StlSpecificationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_stlSpecification)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 8
            self.phi(0)
            self.state = 9
            self.match(stlParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PhiContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return stlParser.RULE_phi

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class PredicateExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def predicate(self):
            return self.getTypedRuleContext(stlParser.PredicateContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPredicateExpr" ):
                listener.enterPredicateExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPredicateExpr" ):
                listener.exitPredicateExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPredicateExpr" ):
                return visitor.visitPredicateExpr(self)
            else:
                return visitor.visitChildren(self)


    class OpFutureExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def FUTUREOP(self):
            return self.getToken(stlParser.FUTUREOP, 0)
        def phi(self):
            return self.getTypedRuleContext(stlParser.PhiContext,0)

        def interval(self):
            return self.getTypedRuleContext(stlParser.IntervalContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOpFutureExpr" ):
                listener.enterOpFutureExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOpFutureExpr" ):
                listener.exitOpFutureExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOpFutureExpr" ):
                return visitor.visitOpFutureExpr(self)
            else:
                return visitor.visitChildren(self)


    class ParenPhiExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LPAREN(self):
            return self.getToken(stlParser.LPAREN, 0)
        def phi(self):
            return self.getTypedRuleContext(stlParser.PhiContext,0)

        def RPAREN(self):
            return self.getToken(stlParser.RPAREN, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterParenPhiExpr" ):
                listener.enterParenPhiExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitParenPhiExpr" ):
                listener.exitParenPhiExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitParenPhiExpr" ):
                return visitor.visitParenPhiExpr(self)
            else:
                return visitor.visitChildren(self)


    class OpUntilExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def phi(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(stlParser.PhiContext)
            else:
                return self.getTypedRuleContext(stlParser.PhiContext,i)

        def UNTILOP(self):
            return self.getToken(stlParser.UNTILOP, 0)
        def interval(self):
            return self.getTypedRuleContext(stlParser.IntervalContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOpUntilExpr" ):
                listener.enterOpUntilExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOpUntilExpr" ):
                listener.exitOpUntilExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOpUntilExpr" ):
                return visitor.visitOpUntilExpr(self)
            else:
                return visitor.visitChildren(self)


    class OpGloballyExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def GLOBALLYOP(self):
            return self.getToken(stlParser.GLOBALLYOP, 0)
        def phi(self):
            return self.getTypedRuleContext(stlParser.PhiContext,0)

        def interval(self):
            return self.getTypedRuleContext(stlParser.IntervalContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOpGloballyExpr" ):
                listener.enterOpGloballyExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOpGloballyExpr" ):
                listener.exitOpGloballyExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOpGloballyExpr" ):
                return visitor.visitOpGloballyExpr(self)
            else:
                return visitor.visitChildren(self)


    class OpLogicalExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def phi(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(stlParser.PhiContext)
            else:
                return self.getTypedRuleContext(stlParser.PhiContext,i)

        def ANDOP(self):
            return self.getToken(stlParser.ANDOP, 0)
        def OROP(self):
            return self.getToken(stlParser.OROP, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOpLogicalExpr" ):
                listener.enterOpLogicalExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOpLogicalExpr" ):
                listener.exitOpLogicalExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOpLogicalExpr" ):
                return visitor.visitOpLogicalExpr(self)
            else:
                return visitor.visitChildren(self)


    class OpReleaseExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def phi(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(stlParser.PhiContext)
            else:
                return self.getTypedRuleContext(stlParser.PhiContext,i)

        def RELEASEOP(self):
            return self.getToken(stlParser.RELEASEOP, 0)
        def interval(self):
            return self.getTypedRuleContext(stlParser.IntervalContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOpReleaseExpr" ):
                listener.enterOpReleaseExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOpReleaseExpr" ):
                listener.exitOpReleaseExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOpReleaseExpr" ):
                return visitor.visitOpReleaseExpr(self)
            else:
                return visitor.visitChildren(self)


    class OpNextExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NEXTOP(self):
            return self.getToken(stlParser.NEXTOP, 0)
        def phi(self):
            return self.getTypedRuleContext(stlParser.PhiContext,0)

        def interval(self):
            return self.getTypedRuleContext(stlParser.IntervalContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOpNextExpr" ):
                listener.enterOpNextExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOpNextExpr" ):
                listener.exitOpNextExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOpNextExpr" ):
                return visitor.visitOpNextExpr(self)
            else:
                return visitor.visitChildren(self)


    class OpPropExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def phi(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(stlParser.PhiContext)
            else:
                return self.getTypedRuleContext(stlParser.PhiContext,i)

        def IMPLIESOP(self):
            return self.getToken(stlParser.IMPLIESOP, 0)
        def EQUIVOP(self):
            return self.getToken(stlParser.EQUIVOP, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOpPropExpr" ):
                listener.enterOpPropExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOpPropExpr" ):
                listener.exitOpPropExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOpPropExpr" ):
                return visitor.visitOpPropExpr(self)
            else:
                return visitor.visitChildren(self)


    class OpNegExprContext(PhiContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a stlParser.PhiContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NEGATION(self):
            return self.getToken(stlParser.NEGATION, 0)
        def phi(self):
            return self.getTypedRuleContext(stlParser.PhiContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOpNegExpr" ):
                listener.enterOpNegExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOpNegExpr" ):
                listener.exitOpNegExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOpNegExpr" ):
                return visitor.visitOpNegExpr(self)
            else:
                return visitor.visitChildren(self)



    def phi(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = stlParser.PhiContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 2
        self.enterRecursionRule(localctx, 2, self.RULE_phi, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 34
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [stlParser.LPAREN]:
                localctx = stlParser.ParenPhiExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 12
                self.match(stlParser.LPAREN)
                self.state = 13
                self.phi(0)
                self.state = 14
                self.match(stlParser.RPAREN)
                pass
            elif token in [stlParser.NEGATION]:
                localctx = stlParser.OpNegExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 16
                self.match(stlParser.NEGATION)
                self.state = 17
                self.phi(9)
                pass
            elif token in [stlParser.NEXTOP]:
                localctx = stlParser.OpNextExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 18
                self.match(stlParser.NEXTOP)
                self.state = 20
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
                if la_ == 1:
                    self.state = 19
                    self.interval()


                self.state = 22
                self.phi(8)
                pass
            elif token in [stlParser.FUTUREOP]:
                localctx = stlParser.OpFutureExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 23
                self.match(stlParser.FUTUREOP)
                self.state = 25
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
                if la_ == 1:
                    self.state = 24
                    self.interval()


                self.state = 27
                self.phi(7)
                pass
            elif token in [stlParser.GLOBALLYOP]:
                localctx = stlParser.OpGloballyExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 28
                self.match(stlParser.GLOBALLYOP)
                self.state = 30
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
                if la_ == 1:
                    self.state = 29
                    self.interval()


                self.state = 32
                self.phi(6)
                pass
            elif token in [stlParser.MINUS, stlParser.NAME]:
                localctx = stlParser.PredicateExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 33
                self.predicate()
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 56
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,7,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 54
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,6,self._ctx)
                    if la_ == 1:
                        localctx = stlParser.OpUntilExprContext(self, stlParser.PhiContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_phi)
                        self.state = 36
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
                        self.state = 37
                        self.match(stlParser.UNTILOP)
                        self.state = 39
                        self._errHandler.sync(self)
                        la_ = self._interp.adaptivePredict(self._input,4,self._ctx)
                        if la_ == 1:
                            self.state = 38
                            self.interval()


                        self.state = 41
                        self.phi(6)
                        pass

                    elif la_ == 2:
                        localctx = stlParser.OpReleaseExprContext(self, stlParser.PhiContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_phi)
                        self.state = 42
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 4)")
                        self.state = 43
                        self.match(stlParser.RELEASEOP)
                        self.state = 45
                        self._errHandler.sync(self)
                        la_ = self._interp.adaptivePredict(self._input,5,self._ctx)
                        if la_ == 1:
                            self.state = 44
                            self.interval()


                        self.state = 47
                        self.phi(5)
                        pass

                    elif la_ == 3:
                        localctx = stlParser.OpLogicalExprContext(self, stlParser.PhiContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_phi)
                        self.state = 48
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                        self.state = 49
                        _la = self._input.LA(1)
                        if not(_la==stlParser.ANDOP or _la==stlParser.OROP):
                            self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 50
                        self.phi(4)
                        pass

                    elif la_ == 4:
                        localctx = stlParser.OpPropExprContext(self, stlParser.PhiContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_phi)
                        self.state = 51
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 52
                        _la = self._input.LA(1)
                        if not(_la==stlParser.IMPLIESOP or _la==stlParser.EQUIVOP):
                            self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 53
                        self.phi(3)
                        pass

             
                self.state = 58
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,7,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx


    class PredicateContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(stlParser.NAME, 0)

        def RELOP(self):
            return self.getToken(stlParser.RELOP, 0)

        def NUMBER(self):
            return self.getToken(stlParser.NUMBER, 0)

        def MINUS(self):
            return self.getToken(stlParser.MINUS, 0)

        def getRuleIndex(self):
            return stlParser.RULE_predicate

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPredicate" ):
                listener.enterPredicate(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPredicate" ):
                listener.exitPredicate(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPredicate" ):
                return visitor.visitPredicate(self)
            else:
                return visitor.visitChildren(self)




    def predicate(self):

        localctx = stlParser.PredicateContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_predicate)
        self._la = 0 # Token type
        try:
            self.state = 66
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,9,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 60
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==stlParser.MINUS:
                    self.state = 59
                    self.match(stlParser.MINUS)


                self.state = 62
                self.match(stlParser.NAME)
                self.state = 63
                self.match(stlParser.RELOP)
                self.state = 64
                self.match(stlParser.NUMBER)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 65
                self.match(stlParser.NAME)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class IntervalContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COMMA(self):
            return self.getToken(stlParser.COMMA, 0)

        def LPAREN(self):
            return self.getToken(stlParser.LPAREN, 0)

        def LBRACK(self):
            return self.getToken(stlParser.LBRACK, 0)

        def NUMBER(self, i:int=None):
            if i is None:
                return self.getTokens(stlParser.NUMBER)
            else:
                return self.getToken(stlParser.NUMBER, i)

        def INF(self, i:int=None):
            if i is None:
                return self.getTokens(stlParser.INF)
            else:
                return self.getToken(stlParser.INF, i)

        def RPAREN(self):
            return self.getToken(stlParser.RPAREN, 0)

        def RBRACK(self):
            return self.getToken(stlParser.RBRACK, 0)

        def getRuleIndex(self):
            return stlParser.RULE_interval

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInterval" ):
                listener.enterInterval(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInterval" ):
                listener.exitInterval(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitInterval" ):
                return visitor.visitInterval(self)
            else:
                return visitor.visitChildren(self)




    def interval(self):

        localctx = stlParser.IntervalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_interval)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 68
            _la = self._input.LA(1)
            if not(_la==stlParser.LPAREN or _la==stlParser.LBRACK):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 69
            _la = self._input.LA(1)
            if not(_la==stlParser.INF or _la==stlParser.NUMBER):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 70
            self.match(stlParser.COMMA)
            self.state = 71
            _la = self._input.LA(1)
            if not(_la==stlParser.INF or _la==stlParser.NUMBER):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 72
            _la = self._input.LA(1)
            if not(_la==stlParser.RPAREN or _la==stlParser.RBRACK):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[1] = self.phi_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def phi_sempred(self, localctx:PhiContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 5)
         

            if predIndex == 1:
                return self.precpred(self._ctx, 4)
         

            if predIndex == 2:
                return self.precpred(self._ctx, 3)
         

            if predIndex == 3:
                return self.precpred(self._ctx, 2)
         




